from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Collection, Callable, TYPE_CHECKING

import imageio
import numpy as np
import pyvista as pv

from metalpy.mepa import LinearExecutor, Executor, ParallelProgress
from metalpy.utils.dhash import dhash
from metalpy.utils.file import make_cache_file_path, PathLike
from metalpy.utils.model import convert_model_dict_to_multiblock, pv_ufunc_apply, DataSetLike, DataAssociation, \
    pv_ufunc_map
from metalpy.utils.time import Timer
from .dataset_wrapper import DataSetWrapper, TDataSetWrapper

if TYPE_CHECKING:
    from metalpy.carto.utils.crs import CRSLike

ModelPostprocessor = Callable[[DataSetLike, Path], DataSetLike]


class UniversalDataSet(DataSetWrapper):
    Transforms = {
        pv.DataSet.rotate_x.__name__,
        pv.DataSet.rotate_y.__name__,
        pv.DataSet.rotate_z.__name__,
        pv.DataSet.rotate_vector.__name__,
        pv.DataSet.translate.__name__,
        pv.DataSet.scale.__name__,
        pv.DataSet.flip_x.__name__,
        pv.DataSet.flip_y.__name__,
        pv.DataSet.flip_z.__name__,
        pv.DataSet.flip_normal.__name__,
    }

    def __init__(self, dataset: DataSetLike, *args, **kwargs):
        super().__init__(dataset, *args, **kwargs)

    def warp(
            self: TDataSetWrapper,
            src_crs: 'CRSLike',
            *,
            crs=None,
            query=None,
            inplace=False,
            return_crs=False
    ) -> TDataSetWrapper:
        """转换数据集的坐标系

        Parameters
        ----------
        src_crs
            源坐标系
        crs
            目标坐标系
        query
            按指定模式搜索坐标系
        inplace
            指定转换是否就地操作
        return_crs
            指示时候返回采用的新坐标系，常用与配合`query`，获取查询到的坐标系信息

        Returns
        -------
        result
            转换坐标系后的数据集，如果指定了`return_crs=True`，则额外返回转换后的坐标系信息
        """
        from metalpy.carto.pyvista.crs import warp_dataset

        return self.wrap(warp_dataset(
            self.dataset,
            src_crs,
            crs=crs,
            query=query,
            inplace=inplace,
            return_crs=return_crs
        ))

    @property
    def n_points(self):
        return self._count_size(DataAssociation.Point)

    @property
    def n_cells(self):
        return self._count_size(DataAssociation.Cell)

    @property
    def points(self):
        association = DataAssociation.Point
        points = pv_ufunc_map(self.dataset, association.get)

        return np.vstack(points)

    @points.setter
    def points(self, points):
        index = [0]
        association = DataAssociation.Point

        def func(dataset):
            n_points = association.get_count(dataset)
            end = index[0] + n_points
            association.set(dataset, points[index[0]:end])
            index[0] = end

        pv_ufunc_apply(
            self.dataset,
            func,
            inplace=True
        )

        assert index[0] == len(points), (
            f'It is not allowed to add extra points by setting'
            f' `{type(self).__name__}.points = ...`.'
            f' Consider merging all blocks into a single `PolyData`.'
        )

    @classmethod
    def read(
            cls: type[TDataSetWrapper],
            model_path: PathLike,
            callback: ModelPostprocessor | None = None
    ) -> TDataSetWrapper | pv.DataSet:
        model = _load_model(model_path, callback=callback)
        return cls(model)

    @classmethod
    def read_all(
            cls: type[TDataSetWrapper],
            path,
            progress=True,
            verbose=True,
            cache=True,
            preview: int | bool = False,
            callback: ModelPostprocessor | None = None,
            executor: Executor | None = None
    ) -> TDataSetWrapper | pv.DataSet:
        """从指定路径中读取全部模型文件

        Parameters
        ----------
        path
            模型路径根目录
        progress
            是否显示进度条
        verbose
            是否显示提示信息
        cache
            是否缓存较大的模型到本地缓存路径下
        preview
            指示只载入前若干个模型文件
        callback
            针对每个载入模型的回调处理函数
        executor
            指定用于执行载入任务的执行器

        Returns
        -------
        dataset
            载入的全部模型数据数据

        Notes
        -----
        使用ProcessExecutor等依赖pickle序列化回传结果的执行器时，序列化PyVista对象耗时会远大于读取时间，因此应用场景有限。

        TODO: 寻找使用ThreadExecutor进行加速的可能性。
            可能可以依靠取消GIL后的ThreadExecutor？
            当前多线程读取PyVista对象依然会受限于GIL，导致IO阻塞。
        """
        path = Path(path)
        model_name = path.name
        cache_path = make_cache_file_path(f'models/{dhash(path).hexdigest(6)}/{model_name}.vtm')

        if preview is True:
            preview = 10

        if cache and cache_path.exists():
            timer = Timer()
            with timer:
                ret = pv.read(cache_path)  # TODO: 判断`texture`选项？
            if verbose:
                print(f'Model loaded from `{path}` in {timer}.')
        else:
            model_files = [
                os.path.join(base, f)
                for base, _, files in os.walk(path)
                for f in files
                if _check_is_readable(Path(f))
            ]

            if preview:
                model_files = model_files[:preview]

            if executor is None:
                executor = LinearExecutor()

            if progress is True:
                total = len(model_files)
                progress = executor.progress(total=total)
            elif progress is False:
                progress = None

            with Timer() as timer:
                allocator = executor.arrange(model_files)
                futures = executor.distribute(_load_models_from_directory, cwd=path, paths=allocator, callback=callback,
                                              progress=progress)

                model_dicts = executor.gather(futures)

                ret = pv.MultiBlock()
                for model_dict in model_dicts:
                    convert_model_dict_to_multiblock(model_dict, root=ret)

                if progress is not None:
                    progress.close()

                if cache and timer.elapsed > 5:
                    if verbose:
                        print(f'Saving loaded model to `{cache_path}`...')

                    with timer:
                        ret.save(cache_path, binary=True)
                        if verbose:
                            print(f'Model saved in {timer}.')

        return cls(ret)

    def __getattr__(self, item):
        if item in UniversalDataSet.Transforms:
            def wrapper(*args, **kwargs):
                inplace = kwargs.pop('inplace', False)
                new_dataset = pv_ufunc_apply(
                    self.dataset,
                    lambda dataset: getattr(dataset, item)(*args, **kwargs, inplace=True),
                    inplace=inplace
                )
                if inplace:
                    return self
                else:
                    return self.wrap(new_dataset)

            return wrapper
        else:
            return super().__getattr__(item)

    def _count_size(self, association: DataAssociation):
        count = pv_ufunc_map(self.dataset, association.get_count)
        return sum(count)


__MODEL_FORMATS = set(pv.core.utilities.reader.CLASS_READERS.keys())
__MODEL_FORMATS.difference_update({
    *imageio.config.known_extensions
})


def _check_is_readable(file: Path, excludes: Collection = None):
    suffix = file.suffix
    flag = file.suffix in __MODEL_FORMATS
    if excludes is not None:
        flag &= suffix not in excludes
    return flag


def _load_models_from_directory(
        cwd,
        paths=None,
        callback=None,
        progress: ParallelProgress | None = None,
):
    root_model: dict[str, Any] = {}

    cwd = Path(cwd)
    if paths is None:
        paths = list(cwd.iterdir())

    for path in paths:
        name_parts, sub_model = _check_directory_and_load(
            cwd=cwd,
            child_path=Path(path),
            callback=callback,
            progress=progress
        )

        if sub_model is not None:
            node = root_model
            for part in name_parts[:-1]:
                node = node.setdefault(part, {})
            node[name_parts[-1]] = sub_model

    return root_model


def _load_model(model_path, callback=None) -> DataSetLike:
    model_path = Path(model_path)
    model = pv.read(os.fspath(model_path))

    if callback is not None:
        model = callback(model, model_path)

    return model


def _check_directory_and_load(
        cwd: Path,
        child_path: Path,
        callback=None,
        progress: ParallelProgress | None = None
):
    sub_model = None
    if _check_is_readable(child_path):
        try:
            sub_model = _load_model(child_path, callback=callback)
        except IOError:
            pass

        if progress is not None:
            progress.update(1)

    if sub_model is not None:
        return child_path.relative_to(cwd).parts, sub_model
    else:
        return tuple(), None



import taichi as ti

from metalpy.utils.taichi import ti_kernel
from metalpy.utils.taichi_kernels import ti_use


@ti_kernel
def kernel_matrix_forward(
        receiver_locations: ti.types.ndarray(),
        xn: ti.types.ndarray(),
        yn: ti.types.ndarray(),
        zn: ti.types.ndarray(),
        base_cell_sizes: ti.types.ndarray(),
        susc_model: ti.types.ndarray(),
        Txx: ti.template(),
        Txy: ti.template(),
        Txz: ti.template(),
        Tyy: ti.template(),
        Tyz: ti.template(),
        Tzz: ti.template(),
        mat: ti.types.ndarray(),
        write_to_mat: ti.template(),
        compressed: ti.template(),
):
    # calculates A = I - X @ T, where T is the forward kernel, s.t. T @ m_v = B_v
    # m_v and B_v are both channel first (Array of Structure in taichi)
    # m_v = [Mx1, My1, Mz1, ... Mxn, Myn, Mzn]
    # B_v = [Bx1, By1, Bz1, ... Bxn, Byn, Bzn]
    #     ┌─────────────────────────── nC ─────────────────────────┐
    #     ┌────────────────┬────────────────┬─────┬────────────────┐ ─┐
    #     │ Txx, Txy, Txz, │ Txx, Txy, Txz, │ ... │ Txx, Txy, Txz, │  │
    #     │ Tyx, Tyy, Tyz, │ Tyx, Tyy, Tyz, │ ... │ Tyx, Tyy, Tyz, │  │
    #     │ Tzx, Tzy, Tzz, │ Tzx, Tzy, Tzz, │ ... │ Tzx, Tzy, Tzz, │  │
    #     ├────────────────┼────────────────┼─────┼────────────────┤  │
    # T = │ ...            │ ...            │ ... │ ...            │ nObs
    #     ├────────────────┼────────────────┼─────┼────────────────┤  │
    #     │ Txx, Txy, Txz, │ Txx, Txy, Txz, │ ... │ Txx, Txy, Txz, │  │
    #     │ Tyx, Tyy, Tyz, │ Tyx, Tyy, Tyz, │ ... │ Tyx, Tyy, Tyz, │  │
    #     │ Tzx, Tzy, Tzz, │ Tzx, Tzy, Tzz, │ ... │ Tzx, Tzy, Tzz, │  │
    #     └────────────────┴────────────────┴─────┴────────────────┘ ─┘
    tol1 = 1e-10  # Tolerance 1 for numerical stability over nodes and edges
    tol2 = 1e-4  # Tolerance 2 for numerical stability over nodes and edges

    # number of cells in mesh
    nC = xn.shape[0]
    nObs = receiver_locations.shape[0]

    # base cell dimensions
    min_hx = base_cell_sizes[0]
    min_hy = base_cell_sizes[1]
    min_hz = base_cell_sizes[2]

    dummy = xn[0, 0]
    total = nObs * nC
    if ti.static(compressed):
        total = Txx.shape[0]

    for i in range(total):
        iobs, icell = 0, 0
        if ti.static(not compressed):
            iobs, icell = i // nC, i % nC

        # comp. pos. differences for tne, bsw nodes. Adjust if location within
        # tolerance of a node or edge
        dx1, dx2, dy1, dy2, dz1, dz2 = dummy, dummy, dummy, dummy, dummy, dummy
        ti_use(dx1)
        ti_use(dx2)
        ti_use(dy1)
        ti_use(dy2)
        ti_use(dz1)
        ti_use(dz2)

        if ti.static(not compressed):
            dx1 = xn[icell, 0] - receiver_locations[iobs, 0]
            dx2 = xn[icell, 1] - receiver_locations[iobs, 0]
            dy1 = yn[icell, 0] - receiver_locations[iobs, 1]
            dy2 = yn[icell, 1] - receiver_locations[iobs, 1]
            dz1 = zn[icell, 0] - receiver_locations[iobs, 2]
            dz2 = zn[icell, 1] - receiver_locations[iobs, 2]
        else:
            dx1 = Txx[i]

            if ti.math.isinf(dx1):
                continue

            dx2 = Txy[i]
            dy1 = Txz[i]
            dy2 = Tyy[i]
            dz1 = Tyz[i]
            dz2 = Tzz[i]

        if ti.abs(dx1) / min_hx < tol2:
            dx1 = tol2 * min_hx
        if ti.abs(dx2) / min_hx < tol2:
            dx2 = tol2 * min_hx

        if ti.abs(dy1) / min_hy < tol2:
            dy1 = tol2 * min_hy
        if ti.abs(dy2) / min_hy < tol2:
            dy2 = tol2 * min_hy

        if ti.abs(dz1) / min_hz < tol2:
            dz1 = tol2 * min_hz
        if ti.abs(dz2) / min_hz < tol2:
            dz2 = tol2 * min_hz

        # comp. squared diff
        dx2dx2 = dx2 ** 2.0
        dx1dx1 = dx1 ** 2.0

        dy2dy2 = dy2 ** 2.0
        dy1dy1 = dy1 ** 2.0

        dz2dz2 = dz2 ** 2.0
        dz1dz1 = dz1 ** 2.0

        # 2D radius component squared of corner nodes
        R1 = dy2dy2 + dx2dx2
        R2 = dy2dy2 + dx1dx1
        R3 = dy1dy1 + dx2dx2
        R4 = dy1dy1 + dx1dx1

        # radius to each cell node
        r1 = ti.sqrt(dz2dz2 + R2)
        r2 = ti.sqrt(dz2dz2 + R1)
        r3 = ti.sqrt(dz1dz1 + R1)
        r4 = ti.sqrt(dz1dz1 + R2)
        r5 = ti.sqrt(dz2dz2 + R3)
        r6 = ti.sqrt(dz2dz2 + R4)
        r7 = ti.sqrt(dz1dz1 + R4)
        r8 = ti.sqrt(dz1dz1 + R3)

        # compactify argument calculations
        arg1 = dy2 + dz2 + r1
        arg2 = dx1 + dz2 + r1
        arg3 = dx1 + r1
        arg4 = dy2 + r1
        arg5 = dz2 + r1

        arg6 = dy2 + dz2 + r2
        arg7 = dx2 + dz2 + r2
        arg8 = dx2 + r2
        arg9 = dy2 + r2
        arg10 = dz2 + r2

        arg11 = dy2 + dz1 + r3
        arg12 = dx2 + dz1 + r3
        arg13 = dx2 + r3
        arg14 = dy2 + r3
        arg15 = dz1 + r3

        arg16 = dy2 + dz1 + r4
        arg17 = dx1 + dz1 + r4
        arg18 = dx1 + r4
        arg19 = dy2 + r4
        arg20 = dz1 + r4

        arg21 = dy1 + dz2 + r5
        arg22 = dx2 + dz2 + r5
        arg23 = dx2 + r5
        arg24 = dy1 + r5
        arg25 = dz2 + r5

        arg26 = dy1 + dz2 + r6
        arg27 = dx1 + dz2 + r6
        arg28 = dx1 + r6
        arg29 = dy1 + r6
        arg30 = dz2 + r6

        arg31 = dy1 + dz1 + r7
        arg32 = dx1 + dz1 + r7
        arg33 = dx1 + r7
        arg34 = dy1 + r7
        arg35 = dz1 + r7

        arg36 = dy1 + dz1 + r8
        arg37 = dx2 + dz1 + r8
        arg38 = dx2 + r8
        arg39 = dy1 + r8
        arg40 = dz1 + r8

        # 判断观测点是否在网格内，如果在网格内，则对角线元素+1
        inside = (
            + ti.math.sign(dx1 * dx2)
            + ti.math.sign(dy1 * dy2)
            + ti.math.sign(dz1 * dz2)
            == -3
        )

        txx = (
            + -2 * ti.atan2(dx1, arg1 + tol1)
            - -2 * ti.atan2(dx2, arg6 + tol1)
            + -2 * ti.atan2(dx2, arg11 + tol1)
            - -2 * ti.atan2(dx1, arg16 + tol1)
            + -2 * ti.atan2(dx2, arg21 + tol1)
            - -2 * ti.atan2(dx1, arg26 + tol1)
            + -2 * ti.atan2(dx1, arg31 + tol1)
            - -2 * ti.atan2(dx2, arg36 + tol1)
        ) / -4 / ti.math.pi

        txy = (
            + ti.log(arg5) - ti.log(arg10)
            + ti.log(arg15) - ti.log(arg20)
            + ti.log(arg25) - ti.log(arg30)
            + ti.log(arg35) - ti.log(arg40)
        ) / -4 / ti.math.pi

        txz = (
            + ti.log(arg4) - ti.log(arg9)
            + ti.log(arg14) - ti.log(arg19)
            + ti.log(arg24) - ti.log(arg29)
            + ti.log(arg34) - ti.log(arg39)
        ) / -4 / ti.math.pi

        tyy = (
            + -2 * ti.atan2(dy2, arg2 + tol1)
            - -2 * ti.atan2(dy2, arg7 + tol1)
            + -2 * ti.atan2(dy2, arg12 + tol1)
            - -2 * ti.atan2(dy2, arg17 + tol1)
            + -2 * ti.atan2(dy1, arg22 + tol1)
            - -2 * ti.atan2(dy1, arg27 + tol1)
            + -2 * ti.atan2(dy1, arg32 + tol1)
            - -2 * ti.atan2(dy1, arg37 + tol1)
        ) / -4 / ti.math.pi

        tyz = (
            + ti.log(arg3) - ti.log(arg8)
            + ti.log(arg13) - ti.log(arg18)
            + ti.log(arg23) - ti.log(arg28)
            + ti.log(arg33) - ti.log(arg38)
        ) / -4 / ti.math.pi

        # txx + tyy + tzz = 0, 如果观测点在网格外
        # txx + tyy + tzz = -1, 如果观测点在网格内
        tzz = -inside - txx - tyy

        tyx = txy
        tzx = txz
        tzy = tyz

        neg_sus = -susc_model[icell]

        if ti.static(compressed):
            Txx[i] = neg_sus * txx + inside
            Txy[i] = neg_sus * txy
            Txz[i] = neg_sus * txz
            Tyy[i] = neg_sus * tyy + inside
            Tyz[i] = neg_sus * tyz
            Tzz[i] = neg_sus * tzz + inside
        else:
            if ti.static(write_to_mat):
                mat[iobs * 3 + 0, icell * 3 + 0] = neg_sus * txx + inside
                mat[iobs * 3 + 0, icell * 3 + 1] = neg_sus * txy
                mat[iobs * 3 + 0, icell * 3 + 2] = neg_sus * txz
                mat[iobs * 3 + 1, icell * 3 + 0] = neg_sus * tyx
                mat[iobs * 3 + 1, icell * 3 + 1] = neg_sus * tyy + inside
                mat[iobs * 3 + 1, icell * 3 + 2] = neg_sus * tyz
                mat[iobs * 3 + 2, icell * 3 + 0] = neg_sus * tzx
                mat[iobs * 3 + 2, icell * 3 + 1] = neg_sus * tzy
                mat[iobs * 3 + 2, icell * 3 + 2] = neg_sus * tzz + inside
            else:
                Txx[iobs, icell] = neg_sus * txx + inside
                Txy[iobs, icell] = neg_sus * txy
                Txz[iobs, icell] = neg_sus * txz
                Tyy[iobs, icell] = neg_sus * tyy + inside
                Tyz[iobs, icell] = neg_sus * tyz
                Tzz[iobs, icell] = neg_sus * tzz + inside
